import time
from functools import wraps
from quotexpy.logger import logger as log


def retry(ExceptionToCheck, tries=4, delay=3, backoff=2, logger=None):
    """Retry calling the decorated function using an exponential backoff.

    http://www.saltycrane.com/blog/2009/11/trying-out-retry-decorator-python/
    original from: http://wiki.python.org/moin/PythonDecoratorLibrary#Retry

    :param ExceptionToCheck: the exception to check. may be a tuple of
        exceptions to check
    :type ExceptionToCheck: Exception or tuple
    :param tries: number of times to try (not retry) before giving up
    :type tries: int
    :param delay: initial delay between retries in seconds
    :type delay: int
    :param backoff: backoff multiplier e.g. value of 2 will double the delay
        each retry
    :type backoff: int
    :param logger: logger to use. If None, print
    :type logger: logging.Logger instance
    """

    def deco_retry(f):
        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except ExceptionToCheck as e:
                    msg = "%s, Retrying in %d seconds..." % (str(e), mdelay)
                    if logger:
                        log.warning(msg)
                    else:
                        log.info(msg)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry

    return deco_retry


Retry = retry



"""Module for Quotex Candles websocket object."""

from quotexpy.ws.objects.base import Base


class ListInfoData(Base):
    """Class for Quotex Candles websocket object."""

    def __init__(self):
        super(ListInfoData, self).__init__()
        self.__name = "listInfoData"
        self.listinfodata_dict = {}

    def set(self, win, game_state, profit, id_number):
        self.listinfodata_dict[id_number] = {
            "win": win,
            "profit": profit,
            "game_state": game_state,
        }

    def delete(self, id_number):
        del self.listinfodata_dict[id_number]

    def get(self, id_number):
        return self.listinfodata_dict[id_number]



from __future__ import annotations
import subprocess
from playwright._impl._driver import compute_driver_executable, get_driver_env
from playwright.async_api import BrowserType as AsyncBrowserType
from playwright.sync_api import BrowserType as SyncBrowserType

__version__ = "0.0.0"
__all__ = ["install"]


def install(
    browser_type: SyncBrowserType | AsyncBrowserType,
    *,
    with_deps: bool = False,
) -> bool:
    """install playwright and deps if needed

    Args:
        browser_type (SyncBrowserType | AsyncBrowserType): `BrowserType` object. Example: `p.chrome`
        with_deps (bool, optional): install with dependencies. Defaults to `False`.

    Returns:
        bool: succeeded or failed
    """
    driver_executable = str(compute_driver_executable())
    args = [driver_executable, "install-deps"]
    env = None
    if browser_type:
        args = [driver_executable, "install", browser_type.name]
        env = get_driver_env()
        if with_deps:
            args.append("--with-deps")

    proc = subprocess.run(args, env=env, capture_output=True, text=True)

    return proc.returncode == 0



import json
from quotexpy.ws.channels.base import Base


class SellOption(Base):
    """Class for Quotex sell option websocket channel."""

    name = "sell_option"

    def __call__(self, options_ids):
        """
        :param options_ids: list or int
        """
        if not isinstance(options_ids, list):
            payload = {"ticket": options_ids}
            self.send_websocket_request(f'42["orders/cancel",{json.dumps(payload)}]')
        else:
            for ids in options_ids:
                payload = {"ticket": ids}
                self.send_websocket_request(f'42["orders/cancel",{json.dumps(payload)}]')



import json
from quotexpy.ws.channels.base import Base


class Ssid(Base):
    """Class for Quotex API ssid websocket channel."""

    name = "ssid"

    def __call__(self, ssid):
        """Method to send message to ssid websocket channel.

        :param ssid: The session identifier.
        """
        payload = {
            "session": ssid,
            "isDemo": self.api.account_type,
            "tournamentId": 0,
        }
        data = f'42["authorization",{json.dumps(payload)}]'
        self.send_websocket_request(data)



import logging
import logging.handlers

logger = logging.getLogger("papertrail")
logger.setLevel(logging.INFO)
papertrail_handler = logging.handlers.SysLogHandler(address=("logs6.papertrailapp.com", 38366))
logger.addHandler(papertrail_handler)



"""Module for Quotex API constants."""

codes_asset = {
    "AUDCAD": 36,
    "AUDCAD_otc": 67,
    "AUDCHF": 37,
    "AUDCHF_otc": 68,
    "AUDJPY": 38,
    "AUDJPY_otc": 69,
    "AUDUSD": 40,
    "AUDUSD_otc": 71,
    "AXJAUD": 315,
    "AXP_otc": 291,
    "BA_otc": 292,
    "CADCHF": 41,
    "CADJPY_otc": 73,
    "CHIA50": 328,
    "DJIUSD": 317,
    "EURAUD": 44,
    "EURCAD": 45,
    "EURCHF": 46,
    "EURCHF_otc": 77,
    "EURGBP": 47,
    "EURGBP_otc": 78,
    "EURJPY": 48,
    "EURJPY_otc": 79,
    "EURSGD": 123,
    "EURUSD": 1,
    "EURUSD_otc": 66,
    "F40EUR": 318,
    "FB_otc": 187,
    "FTSGBP": 319,
    "GBPAUD": 51,
    "GBPAUD_otc": 81,
    "GBPCAD": 52,
    "GBPCHF": 53,
    "GBPJPY": 54,
    "GBPJPY_otc": 84,
    "GBPUSD": 56,
    "GEREUR": 316,
    "HSIHKD": 320,
    "IBXEUR": 321,
    "INTC_otc": 190,
    "IT4EUR": 326,
    "JNJ_otc": 296,
    "JPXJPY": 327,
    "MCD_otc": 175,
    "MSFT_otc": 176,
    "NDXUSD": 322,
    "NZDJPY_otc": 89,
    "NZDUSD": 60,
    "NZDUSD_otc": 90,
    "PFE_otc": 297,
    "SPXUSD": 323,
    "STXEUR": 325,
    "UKBrent_otc": 164,
    "USCrude_otc": 165,
    "USDCAD": 61,
    "USDCAD_otc": 91,
    "USDCHF": 62,
    "USDCHF_otc": 92,
    "USDJPY": 63,
    "USDJPY_otc": 93,
    "XAGUSD": 65,
    "XAGUSD_otc": 167,
    "XAUUSD": 2,
    "XAUUSD_otc": 169,
}



"""Module for Quotex http login resource."""

from quotexpy.http.navigator import Browser


class Logout(Browser):
    """Class for Quotex login resource."""

    url = ""

    def _post(self, data=None, headers=None):
        """Send get request for Quotex API login http resource.
        :returns: The instance of :class:`navigator.Session`.
        """
        return self.send_request(
            method="POST",
            url="https://quotex.com/logout",
            data=data,
            headers=headers,
        )

    def __call__(self):
        return self._post()



"""Module for Quotex websocket."""
import os
import json
import random
import logging
import websocket
from quotexpy import global_value
from quotexpy.http.user_agents import agents

user_agent_list = agents.split("\n")


class WebsocketClient(object):
    """Class for work with Quotex API websocket."""

    def __init__(self, api):
        """
        :param api: The instance of :class:`QuotexAPI
            <quotexapi.api.QuotexAPI>`.
        :trace_ws: Enables and disable `enableTrace` in WebSocket Client.
        """
        self.api = api
        self.headers = {
            "User-Agent": self.api.user_agent
            if not None
            else user_agent_list[random.randint(0, len(user_agent_list) - 1)],
        }
        websocket.enableTrace(self.api.trace_ws)
        self.wss = websocket.WebSocketApp(
            self.api.wss_url,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
            on_open=self.on_open,
            on_pong=self.on_pong,
            header=self.headers,
            cookie=self.api.cookies,
        )

    def on_message(self, wss, message):
        """Method to process websocket messages."""
        global_value.ssl_Mutual_exclusion = True
        try:
            logger = logging.getLogger(__name__)
            message = message
            if "authorization/reject" in str(message):
                if os.path.isfile("session.json"):
                    os.remove("session.json")
                global_value.SSID = None
                global_value.check_rejected_connection = 1
            elif "s_authorization" in str(message):
                global_value.check_accepted_connection = 1
            try:
                message = message[1:]
                message = message.decode()
                logger.debug(message)
                message = json.loads(str(message))
                self.api.profile.msg = message
                if "call" in str(message) or "put" in str(message):
                    self.api.instruments = message
                elif "signals" in str(message):
                    for i in message["signals"]:
                        self.api.signal_data[i[0]][i[2]]["dir"] = i[1][0]["signal"]
                        self.api.signal_data[i[0]][i[2]]["duration"] = i[1][0]["timeFrame"]
                elif message.get("liveBalance") or message.get("demoBalance"):
                    self.api.account_balance = message
                elif message.get("index"):
                    # logger.info(message)
                    self.api.candles.candles_data = message
                elif message.get("id"):
                    self.api.buy_successful[message["asset"]] = message
                    self.api.buy_id[message["asset"]] = message["id"]
                    self.api.timesync.server_timestamp = message["closeTimestamp"]
                elif message.get("ticket"):
                    self.api.sold_options_respond = message
                if message.get("deals"):
                    for get_m in message["deals"]:
                        self.api.profit_in_operation = get_m["profit"]
                        get_m["win"] = True if get_m["profit"] > 0 else False
                        get_m["game_state"] = 1
                        self.api.listinfodata.set(
                            get_m["win"],
                            get_m["game_state"],
                            get_m["profit"],
                            get_m["id"],
                        )
                if message.get("isDemo") and message.get("balance"):
                    self.api.training_balance_edit_request = message
                elif message.get("error"):
                    # logger.info(message)
                    pass
            except:
                pass
            if "51-" in str(message):
                self.api._temp_status = str(message)
            elif self.api._temp_status == """51-["settings/list",{"_placeholder":true,"num":0}]""":
                self.api.settings_list = message
                self.api._temp_status = ""
            elif self.api._temp_status == """51-["history/list/v2",{"_placeholder":true,"num":0}]""":
                self.api.candle_v2_data[message["asset"]] = message["history"]
            elif len(message[0]) == 4:
                ans = {"time": message[0][1], "price": message[0][2]}
                self.api.realtime_price[message[0][0]].append(ans)
        except:
            pass
        wss.send('42["tick"]')
        global_value.ssl_Mutual_exclusion = False

    @staticmethod
    def on_error(wss, error):
        """Method to process websocket errors."""
        logger = logging.getLogger(__name__)
        logger.error(error)
        global_value.websocket_error_reason = str(error)
        global_value.check_websocket_if_error = True

    def on_open(self, wss):
        """Method to process websocket open."""
        logger = logging.getLogger(__name__)
        logger.debug("Websocket client connected.")
        global_value.check_websocket_if_connect = 1

    @staticmethod
    def on_close(wss, close_status_code, close_msg):
        """Method to process websocket close."""
        logger = logging.getLogger(__name__)
        logger.debug("Websocket connection closed.")
        global_value.check_websocket_if_connect = 0

    @staticmethod
    def on_pong(wss, a):
        """Method to process websocket close."""



import json
from quotexpy.ws.channels.base import Base


class GetCandles(Base):
    """Class for Quotex candles websocket channel."""

    name = "candles"

    def __call__(self, asset_id, offset, period, index):
        """Method to send message to candles websocket chanel.

        :param asset_id: The active/asset identifier.
        :param period: The candle duration (timeframe for the candles).
        :param amount: The number of candles you want to have
        """
        payload = {
            "id": asset_id,
            "index": index,
            "time": period,
            "offset": offset,
        }
        data = f'42["history/load/line",{json.dumps(payload)}]'
        self.send_websocket_request(data)



from quotexpy.ws.objects.base import Base


class Profile(Base):
    """Class for Quotex Profile websocket object."""

    def __init__(self):
        super(Profile, self).__init__()
        self.__name = "profile"
        self.__nick_name = None
        self.__profile_id = None
        self.__avatar = None
        self.__live_balance = None
        self.__demo_balance = None
        self.__msg = None
        self.__currency_code = None
        self.__currency_symbol = None
        self.__profile_level = None
        self.__minimum_amount = None

    @property
    def nick_name(self):
        """Property to get nick_name value.

        :returns: The nick_name value.
        """
        return self.__nick_name

    @nick_name.setter
    def nick_name(self, nick_name):
        """Method to set nick_name value."""
        self.__nick_name = nick_name

    @property
    def live_balance(self):
        """Property to get live_balance value.

        :returns: The live_balance value.
        """
        return self.__live_balance

    @live_balance.setter
    def live_balance(self, live_balance):
        """Method to set live_balance value."""
        self.__live_balance = live_balance

    @property
    def profile_id(self):
        """Property to get profile value.

        :returns: The profile value.
        """
        return self.__profile_id

    @profile_id.setter
    def profile_id(self, profile_id):
        """Method to set profile value."""
        self.__profile_id = profile_id

    @property
    def demo_balance(self):
        """Property to get demo_balance value.

        :returns: The demo_balance value.
        """
        return self.__demo_balance

    @demo_balance.setter
    def demo_balance(self, demo_balance):
        """Method to set demo_balance value."""
        self.__demo_balance = demo_balance

    @property
    def avatar(self):
        """Property to get avatar value.

        :returns: The avatar value.
        """
        return self.__avatar

    @avatar.setter
    def avatar(self, avatar):
        """Method to set avatar value."""
        self.__avatar = avatar

    @property
    def msg(self):
        return self.__msg

    @msg.setter
    def msg(self, msg):
        self.__msg = msg

    @property
    def currency_symbol(self):
        return self.__currency_symbol

    @currency_symbol.setter
    def currency_symbol(self, currency_symbol):
        self.__currency_symbol = currency_symbol

    @property
    def minimum_amount(self):
        return self.__minimum_amount

    @property
    def currency_code(self):
        return self.__currency_code

    @currency_code.setter
    def currency_code(self, currency_code):
        self.__currency_code = currency_code
        if self.__currency_code.upper() == "BRL":
            self.__minimum_amount = 2

    @property
    def profile_level(self):
        return self.__profile_level

    @profile_level.setter
    def profile_level(self, profile_level):
        self.__profile_level = profile_level



# python
check_websocket_if_connect = None
# try fix ssl.SSLEOFError: EOF occurred in violation of protocol (_ssl.c:2361)
ssl_Mutual_exclusion = False  # mutex read write
# if false websocket can sent self.websocket.send(data)
# else can not sent self.websocket.send(data)
ssl_Mutual_exclusion_write = False  # if thread wirite
SSID = None
check_rejected_connection = False
check_accepted_connection = False
check_websocket_if_error = False
websocket_error_reason = None
balance_id = None
session = None



import json
from quotexpy.ws.channels.base import Base


class Buy(Base):
    """Class for Quotex buy websocket channel."""

    name = "buy"

    def __call__(self, price, asset, direction, duration, request_id):
        """exp, idx = get_expiration_time(
        int(self.api.timesync.server_timestamp), duration)"""
        # option_type = 100
        # option_type = 1
        # if "_otc" not in asset:
        #     option_type = 1
        # duration = get_expiration_time(
        #     int(self.api.timesync.server_timestamp), duration)
        # payload = {
        #     "chartId": "graph",
        #     "settings": {
        #         "chartId": "graph",
        #         "chartType": 2,
        #         "currentExpirationTime": duration,
        #         "isFastOption": False,
        #         "isFastAmountOption": False,
        #         "isIndicatorsMinimized": False,
        #         "isIndicatorsShowing": True,
        #         "isShortBetElement": False,
        #         "chartPeriod": 4,
        #         "currentAsset": {
        #             "symbol": asset
        #         },
        #         "dealValue": 5,
        #         "dealPercentValue": 1,
        #         "isVisible": True,
        #         "timePeriod": 30,
        #         "gridOpacity": 8,
        #         "isAutoScrolling": 1,
        #         "isOneClickTrade": True,
        #         "upColor": "#0FAF59",
        #         "downColor": "#FF6251"
        #     }
        # }
        # data = f'42["settings/store",{json.dumps(payload)}]'
        # self.send_websocket_request(data)

        # payload = {
        #     "asset": asset,
        #     "amount": price,
        #     "time": duration,
        #     "action": direction,
        #     "isDemo": self.api.account_type,
        #     "tournamentId": 0,
        #     "requestId": request_id,
        #     "optionType": 1
        # }
        # data = f'42["orders/open",{json.dumps(payload)}]'
        # self.send_websocket_request(data)
        self.send_websocket_request(
            f'42["orders/open",{json.dumps({"asset": asset,"amount": price,"time": duration,"action": direction,"isDemo": self.api.account_type,"tournamentId": 0,"requestId": request_id,"optionType": 1})}]'
        )


"""
42["pending/create",{"openType":1,"asset":"EURUSD","openPrice":"1.06401","timeframe":120,"command":"call","amount":5}]
{"pending":{"ticket":"4f296304-2dee-4271-8893-3bc4adc56daf","openType":1,"amount":5,"uid":20522712,"isDemo":1,"asset":"EURUSD","openPrice":"1.06401","timeframe":120,"command":0,"currency":"BRL"}}
42["instruments/follow",{"amount":5,"command":0,"currency":"BRL","isDemo":1,"min_payout":0,"open_price":"1.06401","open_type":1,"symbol":"EURUSD","ticket":"4f296304-2dee-4271-8893-3bc4adc56daf","timeframe":120,"uid":20522712}]
42["settings/store",{"chartId":"graph","settings":{"chartId":"graph","chartType":2,"currentExpirationTime":1677005460,"isFastOption":False,"isFastAmountOption":False,"isIndicatorsMinimized":False,"isIndicatorsShowing":True,"isShortBetElement":False,"chartPeriod":4,"currentAsset":{"symbol":"EURUSD"},"dealValue":5,"dealPercentValue":1,"isVisible":True,"timePeriod":60,"gridOpacity":8,"isAutoScrolling":1,"isOneClickTrade":True,"upColor":"#0FAF59","downColor":"#FF6251"}}]	
"""



class Base(object):
    """Class for base Quotex websocket channel."""

    def __init__(self, api):
        """
        :param api: The instance of :class:`QuotexAPI
            <quotexapi.api.QuotexAPI>`.
        """
        self.api = api

    def send_websocket_request(self, data):
        """Send request to Quotex server websocket.
        :param str data: The websocket channel data.
        :returns: The instance of :class:`requests.Response`.
        """
        return self.api.send_websocket_request(data)



"""Module for Quotex websocket."""
import os
import time
import json
import ssl
import threading
import collections
import urllib3
import certifi
from quotexpy import global_value
from quotexpy.http.login import Login
from quotexpy.http.logout import Logout
from quotexpy.ws.channels.ssid import Ssid
from quotexpy.ws.channels.buy import Buy
from quotexpy.ws.channels.candles import GetCandles
from quotexpy.ws.channels.sell_option import SellOption
from quotexpy.ws.objects.timesync import TimeSync
from quotexpy.ws.objects.candles import Candles
from quotexpy.ws.objects.profile import Profile
from quotexpy.ws.objects.listinfodata import ListInfoData
from quotexpy.ws.client import WebsocketClient
from collections import defaultdict
from quotexpy.logger import logger


def nested_dict(n, typeof):
    if n == 1:
        return defaultdict(typeof)
    return defaultdict(lambda: nested_dict(n - 1, typeof))


urllib3.disable_warnings()

cert_path = certifi.where()
os.environ["SSL_CERT_FILE"] = cert_path
os.environ["WEBSOCKET_CLIENT_CA_BUNDLE"] = cert_path
cacert = os.environ.get("WEBSOCKET_CLIENT_CA_BUNDLE")


class QuotexAPI(object):
    """Class for communication with Quotex API."""

    socket_option_opened = {}
    buy_id = {}
    trace_ws = False
    buy_expiration = None
    current_asset = None
    buy_successful = {}
    account_balance = None
    account_type = None
    instruments = None
    training_balance_edit_request = None
    profit_in_operation = None
    sold_options_respond = None
    sold_digital_options_respond = None
    listinfodata = ListInfoData()
    timesync = TimeSync()
    candles = Candles()

    def __init__(self, host, username, password, proxies=None, browser=False):
        """
        :param str host: The hostname or ip address of a Quotex server.
        :param str username: The username of a Quotex server.
        :param str password: The password of a Quotex server.
        :param proxies: The proxies of a Quotex server.
        """
        self._temp_status = ""
        self.settings_list = {}
        self.username = username
        self.password = password
        self.signal_data = nested_dict(2, dict)
        self.getcandle_data = {}
        self.candle_v2_data = {}
        self.cookies = None
        self.profile = None
        self.websocket_thread = None
        self.wss_url = f"wss://ws2.{host}/socket.io/?EIO=3&transport=websocket"
        self.websocket_client = None
        self.set_ssid = None
        self.user_agent = None
        self.token_login2fa = None
        self.proxies = proxies
        self.browser = browser
        self.realtime_price = {}
        self.profile = Profile()

    @property
    def websocket(self):
        """Property to get websocket.

        :returns: The instance of :class:`WebSocket <websocket.WebSocket>`.
        """
        return self.websocket_client.wss

    def get_candle_v2(self):
        payload = {"_placeholder": True, "num": 0}
        data = f'451-["history/list/v2", {json.dumps(payload)}]'
        return self.send_websocket_request(data)

    def subscribe_realtime_candle(self, asset, size, period):
        self.realtime_price[asset] = collections.deque([], size)
        payload = {"asset": asset, "period": period}
        data = f'42["instruments/update", {json.dumps(payload)}]'
        return self.send_websocket_request(data)

    def unsubscribe_realtime_candle(self, asset):
        data = f'42["subfor", {json.dumps(asset)}]'
        return self.send_websocket_request(data)

    @property
    def logout(self):
        """Property for get Quotex http login resource.
        :returns: The instance of :class:`Login
            <quotexpy.http.login.Login>`.
        """
        return Logout(self)

    @property
    def login(self):
        """Property for get Quotex http login resource.
        :returns: The instance of :class:`Login
            <quotexpy.http.login.Login>`.
        """
        return Login(self)

    @property
    def ssid(self):
        """Property for get Quotex websocket ssid channel.
        :returns: The instance of :class:`Ssid
            <quotexpy.ws.channels.ssid.Ssid>`.
        """
        return Ssid(self)

    @property
    def buy(self):
        """Property for get Quotex websocket ssid channel.
        :returns: The instance of :class:`Buy
            <quotexpy.ws.channels.buy.Buy>`.
        """
        return Buy(self)

    @property
    def sell_option(self):
        return SellOption(self)

    @property
    def getcandles(self):
        """Property for get Quotex websocket candles channel.

        :returns: The instance of :class:`GetCandles
            <quotexpy.ws.channels.candles.GetCandles>`.
        """
        return GetCandles(self)

    def check_session(self):
        data = {}
        if os.path.isfile("session.json"):
            with open("session.json") as file:
                data = json.loads(file.read())
            self.user_agent = data.get("user_agent")
        return data.get("ssid"), data.get("cookies")
        # if global_value.session:
        #     data = global_value.session
        #     self.user_agent = data.get("user_agent")
        # return data.get("ssid"), data.get("cookies")

    def send_websocket_request(self, data, no_force_send=True):
        """Send websocket request to Quotex server.
        :param str data: The websocket request data.
        :param bool no_force_send: Default None.
        """
        # while (global_value.ssl_Mutual_exclusion or global_value.ssl_Mutual_exclusion_write) and no_force_send:
        #     pass
        global_value.ssl_Mutual_exclusion_write = True
        # self.websocket.send('42["tick"]')
        # self.websocket.send('42["indicator/list"]')
        # self.websocket.send('42["drawing/load"]')
        # self.websocket.send('42["pending/list"]')
        # self.websocket.send('42["instruments/update",{"asset":"%s","period":60}]' % self.current_asset)
        # self.websocket.send('42["chart_notification/get"]')
        # self.websocket.send('42["depth/follow","%s"]' % self.current_asset)
        self.websocket.send(data)
        logger.debug(data)
        global_value.ssl_Mutual_exclusion_write = False

    def edit_training_balance(self, amount):
        data = f'42["demo/refill",{json.dumps(amount)}]'
        self.send_websocket_request(data)

    async def get_ssid(self):
        ssid, cookies = self.check_session()
        if not ssid:
            # try:
            logger.info("Authenticating user...")
            ssid, cookies = await self.login(
                self.username,
                self.password,
                self.browser,
            )
            logger.info("Login successful!!!")
            # except Exception as e:
            # logger = logging.getLogger(__name__)
            # logger.error(e)
            # return e
        return ssid, cookies

    def start_websocket(self):
        global_value.check_websocket_if_connect = None
        global_value.check_websocket_if_error = False
        global_value.websocket_error_reason = None
        self.websocket_client = WebsocketClient(self)
        self.websocket_thread = threading.Thread(
            target=self.websocket.run_forever,
            kwargs={
                "ping_interval": 2,
                "ping_payload": "2",
                "origin": "https://qxbroker.com",
                "host": "ws2.qxbroker.com",
                "sslopt": {
                    "check_hostname": False,
                    "cert_reqs": ssl.CERT_NONE,
                    "ca_certs": cacert,
                },
            },
        )
        self.websocket_thread.daemon = True
        self.websocket_thread.start()
        while True:
            try:
                if global_value.check_websocket_if_error:
                    return False, global_value.websocket_error_reason
                if global_value.check_websocket_if_connect == 0:
                    logger.info("Websocket connection closed.")
                    logger.debug("Websocket connection closed.")
                    return False, "Websocket connection closed."
                if global_value.check_websocket_if_connect == 1:
                    logger.debug("Websocket successfully connected!!!")
                    return True, "Websocket successfully connected!!!"
            except:
                pass

    def send_ssid(self):
        self.profile.msg = None
        self.ssid(global_value.SSID)
        count = 0
        while not self.profile.msg:
            time.sleep(0.1)
            count += 1
            if count == 50:
                raise Exception(f"Sending authorization with SSID {global_value.SSID} took too long to respond")
        if not self.profile.msg:
            return False
        return True

    async def connect(self):
        """Method for connection to Quotex API."""
        global_value.ssl_Mutual_exclusion = False
        global_value.ssl_Mutual_exclusion_write = False
        if global_value.check_websocket_if_connect:
            self.close()
        ssid, self.cookies = await self.get_ssid()
        check_websocket, websocket_reason = self.start_websocket()
        if not check_websocket:
            return check_websocket, websocket_reason
        if not global_value.SSID:
            global_value.SSID = ssid
        return check_websocket, websocket_reason

    def close(self):
        if self.websocket_client:
            self.websocket.close()
            self.websocket_thread.join()

    def websocket_alive(self):
        return self.websocket_thread.is_alive()



import sys
import math
import time
import logging
from collections import defaultdict
from quotexpy import expiration
from quotexpy import global_value
from quotexpy.api import QuotexAPI
from quotexpy.constants import codes_asset


def nested_dict(n, typeof):
    if n == 1:
        return defaultdict(typeof)
    return defaultdict(lambda: nested_dict(n - 1, typeof))


def truncate(f, n):
    return math.floor(f * 10**n) / 10**n


class Quotex(object):
    __version__ = "1.0"

    def __init__(self, email, password, browser=False):
        self.size = [
            1,
            5,
            10,
            15,
            30,
            60,
            120,
            300,
            600,
            900,
            1800,
            3600,
            7200,
            14400,
            28800,
            43200,
            86400,
            604800,
            2592000,
        ]
        self.email = email
        self.password = password
        self.browser = browser
        self.set_ssid = None
        self.duration = None
        self.suspend = 0.5
        self.subscribe_candle = []
        self.subscribe_candle_all_size = []
        self.subscribe_mood = []
        self.websocket_client = None
        self.websocket_thread = None
        self.debug_ws_enable = False
        self.api = None

    @property
    def websocket(self):
        """Property to get websocket.

        :returns: The instance of :class:`WebSocket <websocket.WebSocket>`.
        """
        return self.websocket_client.wss

    @staticmethod
    def check_connect():
        if global_value.check_websocket_if_connect == 0:
            return False
        return True

    def re_subscribe_stream(self):
        try:
            for ac in self.subscribe_candle:
                sp = ac.split(",")
                self.start_candles_one_stream(sp[0], sp[1])
        except Exception:
            pass
        try:
            for ac in self.subscribe_candle_all_size:
                self.start_candles_all_size_stream(ac)
        except Exception:
            pass
        try:
            for ac in self.subscribe_mood:
                self.start_mood_stream(ac)
        except Exception:
            pass

    def get_instruments(self):
        time.sleep(self.suspend)
        self.api.instruments = None
        while self.api.instruments is None:
            try:
                self.api.get_instruments()
                start = time.time()
                while self.api.instruments is None and time.time() - start < 10:
                    pass
            except Exception:
                logging.error("**error** api.get_instruments need reconnect")
                self.connect()
        return self.api.instruments

    def get_all_asset_name(self):
        if self.api.instruments:
            return [instrument[2].replace("\n", "") for instrument in self.api.instruments]

    def check_asset_open(self, instrument):
        if self.api.instruments:
            for i in self.api.instruments:
                if instrument == i[2]:
                    return i[0], i[2], i[14]

    def get_candles(self, asset, offset, period=None):
        index = expiration.get_timestamp()
        # index - offset
        if period:
            period = expiration.get_period_time(period)
        else:
            period = index
        self.api.current_asset = asset
        self.api.candles.candles_data = None
        while True:
            try:
                self.api.getcandles(codes_asset[asset], offset, period, index)
                while self.check_connect and self.api.candles.candles_data is None:
                    pass
                if self.api.candles.candles_data is not None:
                    break
            except Exception:
                logging.error("**error** get_candles need reconnect")
                self.connect()
        return self.api.candles.candles_data

    def get_candle_v2(self, asset, period):
        self.api.candle_v2_data[asset] = None
        size = 10
        self.stop_candles_stream(asset)
        self.api.subscribe_realtime_candle(asset, size, period)
        while self.api.candle_v2_data[asset] is None:
            time.sleep(1)
        return self.api.candle_v2_data[asset]

    async def connect(self):
        if global_value.check_websocket_if_connect:
            self.close()
        self.api = QuotexAPI(
            "qxbroker.com",
            self.email,
            self.password,
            browser=self.browser,
        )
        self.api.trace_ws = self.debug_ws_enable
        check, reason = await self.api.connect()
        if check:
            self.api.send_ssid()
            if global_value.check_accepted_connection == 0:
                check, reason = await self.connect()
                if not check:
                    check, reason = (
                        False,
                        "Access denied, session does not exist!!!",
                    )
        return check, reason

    def change_account(self, balance_mode="PRACTICE"):
        """Change active account `real` or `practice`"""
        if balance_mode.upper() == "REAL":
            self.api.account_type = 0
        elif balance_mode.upper() == "PRACTICE":
            self.api.account_type = 1
        else:
            logging.error("ERROR doesn't have this mode")
            sys.exit(1)
        self.api.send_ssid()

    def edit_practice_balance(self, amount=None):
        self.api.training_balance_edit_request = None
        self.api.edit_training_balance(amount)
        while self.api.training_balance_edit_request is None:
            pass
        return self.api.training_balance_edit_request

    def get_balance(self):
        while not self.api.account_balance:
            time.sleep(0.1)
        balance = (
            self.api.account_balance.get("demoBalance")
            if self.api.account_type > 0
            else self.api.account_balance.get("liveBalance")
        )
        # return float(f"{truncate(balance + self.get_profit(), 2):.2f}")
        return round(balance, 2)

    def buy(self, price, asset, direction, duration):
        """Buy Binary option"""
        count = 0
        status_buy = False
        self.duration = duration - 1
        request_id = expiration.get_timestamp()
        self.api.current_asset = asset
        self.api.buy_id[asset] = None
        self.api.buy_successful[asset] = None
        self.api.buy(price, asset, direction, duration, request_id)
        while not self.api.buy_id[asset]:
            if count == 10:
                break
            count += 1
            time.sleep(1)
        else:
            status_buy = True
        return status_buy, self.api.buy_successful[asset]

    def sell_option(self, options_ids):
        """Sell asset Quotex"""
        self.api.sell_option(options_ids)
        self.api.sold_options_respond = None
        while self.api.sold_options_respond is None:
            pass
        return self.api.sold_options_respond

    def get_payment(self):
        """Payment Quotex server"""
        assets_data = {}
        for i in self.api.instruments:
            assets_data[i[2]] = {
                "turbo_payment": i[18],
                "payment": i[5],
                "open": i[14],
            }
        return assets_data

    def check_win(self, id_number):
        """Check win based id"""
        # now_stamp = datetime.fromtimestamp(expiration.get_timestamp())
        # expiration_stamp = datetime.fromtimestamp(self.api.timesync.server_timestamp)
        # remaing_time = int((expiration_stamp - now_stamp).total_seconds())
        while True:
            try:
                listinfodata_dict = self.api.listinfodata.get(id_number)
                if listinfodata_dict["game_state"] == 1:
                    break
            except Exception:
                pass
            # remaing_time -= 1
            # time.sleep(1)
            # logger.info(f"\rRestando {remaing_time} segundos ...", end="")
        self.api.listinfodata.delete(id_number)
        # return listinfodata_dict["win"]
        return listinfodata_dict

    def start_candles_stream(self, asset, size, period=0):
        self.api.subscribe_realtime_candle(asset, size, period)

    def stop_candles_stream(self, asset):
        self.api.unsubscribe_realtime_candle(asset)

    def get_realtime_candles(self, asset):
        while True:
            if asset in self.api.realtime_price:
                if len(self.api.realtime_price[asset]) > 0:
                    return self.api.realtime_price[asset]

    def get_signal_data(self):
        return self.api.signal_data

    def get_profit(self):
        return self.api.profit_in_operation or 0

    def start_candles_one_stream(self, asset, size):
        if str(asset + "," + str(size)) not in self.subscribe_candle:
            self.subscribe_candle.append((asset + "," + str(size)))
        start = time.time()
        self.api.candle_generated_check[str(asset)][int(size)] = {}
        while True:
            if time.time() - start > 20:
                logging.error("**error** start_candles_one_stream late for 20 sec")
                return False
            try:
                if self.api.candle_generated_check[str(asset)][int(size)]:
                    return True
            except Exception:
                pass
            try:
                self.api.subscribe(codes_asset[asset], size)
            except Exception:
                logging.error("**error** start_candles_stream reconnect")
                self.connect()
            time.sleep(1)

    def start_candles_all_size_stream(self, asset):
        self.api.candle_generated_all_size_check[str(asset)] = {}
        if str(asset) not in self.subscribe_candle_all_size:
            self.subscribe_candle_all_size.append(str(asset))
        start = time.time()
        while True:
            if time.time() - start > 20:
                logging.error(f"**error** fail {asset} start_candles_all_size_stream late for 10 sec")
                return False
            try:
                if self.api.candle_generated_all_size_check[str(asset)]:
                    return True
            except Exception:
                pass
            try:
                self.api.subscribe_all_size(codes_asset[asset])
            except Exception:
                logging.error("**error** start_candles_all_size_stream reconnect")
                self.connect()
            time.sleep(1)

    def start_mood_stream(self, asset, instrument="turbo-option"):
        if asset not in self.subscribe_mood:
            self.subscribe_mood.append(asset)
        while True:
            self.api.subscribe_Traders_mood(asset[asset], instrument)
            try:
                self.api.traders_mood[codes_asset[asset]] = codes_asset[asset]
                break
            except Exception:
                time.sleep(5)

    def close(self):
        self.api.close()



import random
import requests
import cloudscraper
from bs4 import BeautifulSoup
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from quotexpy.http.user_agents import agents

retry_strategy = Retry(
    total=3,
    backoff_factor=1,
    status_forcelist=[429, 500, 502, 503, 504, 104],
    allowed_methods=["HEAD", "POST", "PUT", "GET", "OPTIONS"],
)
adapter = HTTPAdapter(max_retries=retry_strategy)
user_agent_list = agents.split("\n")


class Browser(object):
    session = requests.Session()
    session.mount("https://", adapter)
    session.mount("http://", adapter)

    def __init__(self, api):
        self.api = api
        self.response = None
        self.headers = self.get_headers()
        self.api.user_agent = self.headers["User-Agent"]

    def get_headers(self):
        self.headers = {
            "User-Agent": user_agent_list[random.randint(0, len(user_agent_list) - 1)],
        }
        return self.headers

    def get_soup(self):
        return BeautifulSoup(self.response.content, "html.parser")

    def send_request(self, method, url, **kwargs):
        self.response = self.session.request(method, url, headers=self.headers, **kwargs)
        if not self.response:
            self.session = cloudscraper.create_scraper()
            self.response = self.session.request(method, url, headers=self.headers, **kwargs)

        return self.response



import time
import datetime
from quotexpy.ws.objects.base import Base


class TimeSync(Base):
    """Class for Quotex TimeSync websocket object."""

    def __init__(self):
        super(TimeSync, self).__init__()
        self.__name = "timeSync"
        self.__server_timestamp = time.time()
        self.__expiration_time = 1

    @property
    def server_timestamp(self):
        """Property to get server timestamp.

        :returns: The server timestamp.
        """
        while self.__server_timestamp is None:
            time.sleep(0.2)

        return self.__server_timestamp

    @server_timestamp.setter
    def server_timestamp(self, timestamp):
        """Method to set server timestamp."""
        self.__server_timestamp = timestamp

    @property
    def server_datetime(self):
        """Property to get server datetime.

        :returns: The server datetime.
        """
        return datetime.datetime.fromtimestamp(self.server_timestamp)

    @property
    def expiration_time(self):
        """Property to get expiration time.

        :returns: The expiration time.
        """
        return self.__expiration_time

    @expiration_time.setter
    def expiration_time(self, minutes):
        """Method to set expiration time

        :param int minutes: The expiration time in minutes.
        """
        self.__expiration_time = minutes

    @property
    def expiration_datetime(self):
        """Property to get expiration datetime.

        :returns: The expiration datetime.
        """
        return self.server_datetime + datetime.timedelta(minutes=self.expiration_time)

    @property
    def expiration_timestamp(self):
        """Property to get expiration timestamp.

        :returns: The expiration timestamp.
        """
        return time.mktime(self.expiration_datetime.timetuple())



import re
import json
from pathlib import Path
from quotexpy.http.navigator import Browser
from quotexpy.http.qxbroker import authorize


class Login(Browser):
    """Class for Quotex login resource."""

    url = ""
    cookies = None
    ssid = None
    base_url = "qxbroker.com"
    https_base_url = f"https://{base_url}"

    def get_token(self):
        self.response = self.send_request("GET", "https://qxbroker.com/")
        self.headers["referer"] = f"{self.https_base_url}/"
        self.response = self.send_request("GET", f"{self.https_base_url}/pt/sign-in/")
        self.cookies = "; ".join(["%s=%s" % (i.name, i.value) for i in self.session.cookies])
        return self.get_soup().find("input", {"name": "_token"})["value"]

    def awaiting_pin(self, data):
        self.headers["Content-Type"] = "application/x-www-form-urlencoded"
        data["keep_code"] = 1
        data["code"] = int(input("Enter the PIN code we just sent to your email: "))
        self.send_request(method="POST", url=f"{self.https_base_url}/pt/sign-in/", data=data)

    def get_profile(self):
        self.response = self.send_request(method="GET", url=f"{self.https_base_url}/pt/trade")
        if self.response:
            script = self.get_soup().find_all("script", {"type": "text/javascript"})[1].get_text()
            match = re.sub("window.settings = ", "", script.strip().replace(";", ""))
            self.ssid = json.loads(match).get("token")
            output_file = Path("session.json")
            output_file.parent.mkdir(exist_ok=True, parents=True)
            output_file.write_text(
                json.dumps(
                    {
                        "cookies": self.cookies,
                        "ssid": self.ssid,
                        "user_agent": self.api.user_agent,
                    },
                    indent=4,
                )
            )
            # global_value.session = {"cookies": self.cookies, "ssid": self.ssid, "user_agent": self.api.user_agent}
            return self.response, json.loads(match)
        return None, None

    def _get(self):
        return self.send_request(method="GET", url=f"f{self.https_base_url}/pt/trade")

    def _post(self, data):
        """Send get request for Quotex API login http resource.
        :returns: The instance of :class:`requests.Response`.
        """
        self.headers["Content-Type"] = "application/x-www-form-urlencoded"
        self.response = self.send_request(method="POST", url=f"{self.https_base_url}/pt/sign-in/", data=data)
        if "Insira o código PIN que acabamos de enviar para o seu e-mail" in self.get_soup().get_text():
            self.awaiting_pin(data)
        result_data = self.get_profile()
        return result_data

    async def __call__(self, username, password, browser):
        """Method to get Quotex API login http request.
        :param str username: The username of a Quotex server.
        :param str password: The password of a Quotex server.
        :returns: The instance of :class:`requests.Response`.
        """
        if browser:
            self.ssid, self.cookies = await authorize(username, password)
        else:
            data = {
                "_token": self.get_token(),
                "email": username,
                "password": password,
                "remember": 1,
            }
            response, self.api.profile.msg = self._post(data=data)
        return self.ssid, self.cookies



import time
import calendar
from datetime import datetime, timedelta


def get_timestamp():
    return calendar.timegm(time.gmtime())


def date_to_timestamp(dt):
    return time.mktime(dt.timetuple())


def get_expiration_time(timestamp, duration):
    now = datetime.now()
    new_date = now.replace(second=0, microsecond=0)
    exp = new_date + timedelta(seconds=duration)
    exp_date = exp.replace(second=0, microsecond=0)
    return int(date_to_timestamp(exp_date))


def get_period_time(duration):
    now = datetime.now()
    period_date = now - timedelta(seconds=duration)
    return int(date_to_timestamp(period_date))


def get_remaning_time(timestamp):
    now_date = datetime.fromtimestamp(timestamp)
    exp_date = now_date.replace(second=0, microsecond=0)
    if (int(date_to_timestamp(exp_date + timedelta(minutes=1))) - timestamp) > 30:
        exp_date = exp_date + timedelta(minutes=1)
    else:
        exp_date = exp_date + timedelta(minutes=2)
    exp = []
    for _ in range(5):
        exp.append(date_to_timestamp(exp_date))
        exp_date = exp_date + timedelta(minutes=1)
    idx = 11
    index = 0
    now_date = datetime.fromtimestamp(timestamp)
    exp_date = now_date.replace(second=0, microsecond=0)
    while index < idx:
        if int(exp_date.strftime("%M")) % 15 == 0 and (int(date_to_timestamp(exp_date)) - int(timestamp)) > 60 * 5:
            exp.append(date_to_timestamp(exp_date))
            index = index + 1
        exp_date = exp_date + timedelta(minutes=1)

    remaning = []

    for idx, t in enumerate(exp):
        if idx >= 5:
            dr = 15 * (idx - 4)
        else:
            dr = idx + 1
        remaning.append((dr, int(t) - int(time.time())))

    return remaning



import re
import json
import requests
from pathlib import Path
from bs4 import BeautifulSoup
from typing import Tuple, Any
from quotexpy.utils.playwright_install import install
from playwright.async_api import Playwright, async_playwright


async def run(username, password, playwright: Playwright) -> Tuple[Any, str]:
    browser = await playwright.firefox.launch(headless=True)
    context = await browser.new_context()
    page = await context.new_page()
    await page.goto("https://qxbroker.com/pt/sign-in")
    await page.get_by_role("textbox", name="E-mail").click()
    await page.get_by_role("textbox", name="E-mail").fill(username)
    await page.get_by_role("textbox", name="Senha").click()
    await page.get_by_role("textbox", name="Senha").fill(password)
    await page.get_by_role("button", name="Entrar").click()
    await page.wait_for_url("https://qxbroker.com/pt/trade")
    cookies = await context.cookies()
    source = await page.content()
    soup = BeautifulSoup(source, "html.parser")
    user_agent = await page.evaluate("() => navigator.userAgent;")
    script = soup.find_all("script", {"type": "text/javascript"})[1].get_text()
    match = re.sub("window.settings = ", "", script.strip().replace(";", ""))
    ssid = json.loads(match).get("token")
    output_file = Path("session.json")
    output_file.parent.mkdir(exist_ok=True, parents=True)
    cookiejar = requests.utils.cookiejar_from_dict({c["name"]: c["value"] for c in cookies})
    cookie_string = "; ".join([f"{c.name}={c.value}" for c in cookiejar])
    output_file.write_text(
        json.dumps(
            {"cookies": cookie_string, "ssid": ssid, "user_agent": user_agent},
            indent=4,
        )
    )
    # global_value.session = {"cookies": cookie_string, "ssid": ssid, "user_agent": user_agent}
    await context.close()
    await browser.close()

    return ssid, cookie_string


async def main(username, password) -> Tuple[Any, str]:
    async with async_playwright() as playwright:
        install(playwright.firefox, with_deps=True)
        return await run(username, password, playwright)


async def authorize(username, password):
    return await main(username, password)
